/* File: sound.c

	SHARP MZ-2000/2200/80B/B2 Emulator "emz2000"
		Sound Modules

	Copyright (C) 2002-2019 FUKUI, Toshio.
	This file is part of the emz2000 software.
	See copyright notice in the COPYING file.
*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <memory.h>
#include <unistd.h>
#include <pthread.h>
#include <esd.h>

#include "common.h"
#include "mz2000.h"
#include "wavefile.h"

#ifndef _POSIX_THREAD_PRIORITY_SCHEDULING
#error "_POSIX_THREAD_PRIORITY_SCHEDULING is not supported\n"
#endif

#undef ENABLE_DISPBUFFER


/* 8bit output values */
#define SOUND_PULSE8_M		0x80U
#define SOUND_PULSE8_VAL	0x60U
/* 16bit output values */
#define SOUND_PULSE16_M		0x0000U	/* only lower=upper=0, because a byte referenced. */
#define SOUND_PULSE16_VAL	0x6000U

#define SOUND_PLAYBUFFER_DIV	100

/* for esd */
static int esd_available = FALSE;
static int esd_sock = - 1;
static int esd_sock2 = - 1;
static esd_info_t *esd_ctrl = NULL;
static esd_format_t esd_format;
static int esd_rate;
static unsigned char *esd_playbuf = NULL;
static volatile int esd_playindex = 0;
static volatile int sound_writeindex = 0;

/* for thread */
static pthread_t soundthread_id = -1, soundthread2_id = -1;
static volatile int soundthread_stopstatus = TRUE;
static volatile int soundthread_waitstatus = TRUE;
static volatile int soundthread2_stopstatus = TRUE;
static volatile int soundthread2_waitstatus = TRUE;
static pthread_mutex_t sound_mutex;
static int sound_thread_realtime1 = FALSE;
static int sound_thread_realtime2 = FALSE;
static int sound_thread_realtime = FALSE;

/* for playdata - I */
static int sound_volume1;
static int sound_volume2;
static int sound_samplingrate;
static int sound_bytes;
static int sound_delaytime;
static int sound_datasize;
static int sound_bufsize;
static int sound_multiply;
static int sound_playbuffer;
static int sound_playoffset;
static int sound_playleftmax;
static int sound_playleftmin;

/* for playdata - II */
static unsigned char *sound_databuf = NULL;
static unsigned char *sound_attrbuf = NULL;
static unsigned char *sound_playbuf = NULL;
static volatile long long sound_baseclock = 0LL;
static unsigned char sound_playdata8   = SOUND_PULSE8_M;
static unsigned short sound_playdata16 = SOUND_PULSE16_M;
static int sound_lastonoff = 0;
static int sound_lastoffset = 0;

/* for status */
int sound_stat_value1;
int sound_stat_value2;
int sound_stat_value3;
/* for sound time interpolations */
static int sound_stat_oldvalue1 = -1;
static int sound_stat_oldvalue2 = -1;

/* for wavefiles */
typedef struct {
	const char *fn;
	int id;
	int device_type;	/* 1 = TAPE / 2 = FD / 3 = QD / 0 = others */
} SOUND_WAVEDATA;

static SOUND_WAVEDATA sound_wavedata[] = {
	{"fddmotor.wav",	-1,	2},
	{"fddmotor_st.wav",	-1,	2},
	{"fddeject.wav",	-1,	2},
	{"fddseek.wav",		-1,	2},
	{"fddseek1.wav",	-1,	2},
	{"cmteject2000.wav",	-1,	1},
	{"cmtplay2000.wav",	-1,	1},
	{"cmtstop2000.wav",	-1,	1},
	{"cmtff2000.wav",	-1,	1},
	{"cmtrew2000.wav",	-1,	1},
	{"cmtffst2000.wav",	-1,	1},
	{"cmtapss2000.wav",	-1,	1},
	{"cmteject80b.wav",	-1,	1},
	{"cmtplay80b.wav",	-1,	1},
	{"cmtstop80b.wav",	-1,	1},
	{"cmtff80b.wav",	-1,	1},
	{"cmtrew80b.wav",	-1,	1}
};

void sound_reset( MZ2000 *mz )
{
	if (!esd_available)
		return;
	if (!sound_thread_realtime)
		pthread_mutex_lock( &sound_mutex );
	if (sound_bytes == 1) {
		memset( sound_databuf, SOUND_PULSE8_M, sound_bufsize );
		memset( sound_playbuf, SOUND_PULSE8_M, sound_bufsize );
		memset( esd_playbuf, SOUND_PULSE8_M, sound_playbuffer );
	} else {
		memset( sound_databuf, SOUND_PULSE16_M, sound_bufsize );
		memset( sound_playbuf, SOUND_PULSE16_M, sound_bufsize );
		memset( esd_playbuf, SOUND_PULSE16_M, sound_playbuffer );
	}
	memset( sound_attrbuf, 0, sound_bufsize );
	sound_lastoffset = 0;
	sound_playdata8  = SOUND_PULSE8_M;
	sound_playdata16 = SOUND_PULSE16_M;
	sound_baseclock = get_internal_clock();
	sound_writeindex = esd_playindex;
	sound_writeindex = (sound_writeindex + sound_playoffset) % sound_playbuffer;
	if (esd_playindex < sound_writeindex)
		sound_stat_value1 = sound_writeindex - esd_playindex;
	else
		sound_stat_value1 = sound_writeindex + sound_playbuffer - esd_playindex;
	if (!sound_thread_realtime)
		pthread_mutex_unlock( &sound_mutex );
	return;
}

void sound_setstate( MZ2000 *mz, int onoff )
{
	int pos, posmod;
	long long posll, t;

	if (!esd_available)
		return;
	if (!(mz -> soundmode) || mz -> playStop || mz -> pause) {
		sound_reset( mz );
		return;
	}
	if (getspeed_z80() >= DEFAULT_SPEED_KHZ * 20)	/* >= 80MHz */
		return;
	if (!sound_thread_realtime)
		pthread_mutex_lock( &sound_mutex );
	posll = (get_internal_clock() - sound_baseclock) * (sound_datasize / sound_bytes);
	t = get_frameclk_z80() * sound_multiply;
	pos = (int)(posll / t);
	if (sound_bytes == 1) {
		posmod = (int)((posll % t) * SOUND_PULSE8_VAL / t);
		//if (pos >= sound_bufsize - 10)
		//	_TRACE("pos over\n");
		if (pos > sound_bufsize - 10)
			pos = sound_bufsize - 10;
		if (pos < 0)
			pos = 0;
		if (pos < sound_lastoffset)
			sound_lastoffset -= sound_datasize;
	} else {
		posmod = (int)((posll % t) * SOUND_PULSE16_VAL / t);
		//if (pos >= sound_bufsize - 10)
		//	_TRACE("pos over\n");
		if (pos * 2 > sound_bufsize - 10)
			pos = (sound_bufsize - 10) / 2;
		if (pos < 0)
			pos = 0;
		if (pos < sound_lastoffset)
			sound_lastoffset -= sound_datasize / 2;
	}
#if 0
	{
		static tcnt = 0;

		tcnt = (tcnt + 1) % 200;
		if (!tcnt)
			_TRACE("%d: %d %d\n", direct_e2, pos - sound_lastoffset, onoff );
	}
#endif
	if (sound_lastonoff != onoff) {
		if (sound_bytes == 1) {
			if (onoff) {
				sound_databuf[pos] = SOUND_PULSE8_M + SOUND_PULSE8_VAL - posmod * 2;
				sound_attrbuf[pos] = 1;
				sound_databuf[pos + 1] = SOUND_PULSE8_M + SOUND_PULSE8_VAL;
				sound_attrbuf[pos + 1] = 1;
			} else {
				sound_databuf[pos] = SOUND_PULSE8_M - SOUND_PULSE8_VAL + posmod * 2;
				sound_attrbuf[pos] = 1;
				sound_databuf[pos + 1] = SOUND_PULSE8_M - SOUND_PULSE8_VAL;
				sound_attrbuf[pos + 1] = 1;
			}
		} else {
			unsigned short t;

			if (onoff) {
				t = SOUND_PULSE16_VAL - posmod * 2;
				sound_databuf[pos * 2    ] = t & 0xff;
				sound_databuf[pos * 2 + 1] = (t >> 8) & 0xff;
				sound_attrbuf[pos] = 1;
				t = SOUND_PULSE16_VAL;
				sound_databuf[pos * 2 + 2] = t & 0xff;
				sound_databuf[pos * 2 + 3] = (t >> 8) & 0xff;
				sound_attrbuf[pos + 1] = 1;
			} else {
				t = posmod * 2 - SOUND_PULSE16_VAL;
				sound_databuf[pos * 2    ] = t & 0xff;
				sound_databuf[pos * 2 + 1] = (t >> 8) & 0xff;
				sound_attrbuf[pos] = 1;
				t = (unsigned short)(0 - SOUND_PULSE16_VAL);
				sound_databuf[pos * 2 + 2] = t & 0xff;
				sound_databuf[pos * 2 + 3] = (t >> 8) & 0xff;
				sound_attrbuf[pos + 1] = 1;
			}
		}
	}
	sound_lastoffset = pos;
	sound_lastonoff = onoff;
	if (!sound_thread_realtime)
		pthread_mutex_unlock( &sound_mutex );
	return;
}

static void sound_thread( MZ2000 *mz )
{
	int i, j, pos, tpause, ridx, widx;
	int n, posdiff, posnct;
	unsigned char c, c1, c2, *cp;
	long long posll, tclk;

	soundthread_stopstatus = FALSE;
	while (!soundthread_stopstatus) {
		if (!esd_available) {
			usleep( (mz -> sound_playbuffer_ms) * 1000 );
			continue;
		}
		if (!sound_thread_realtime)
			pthread_mutex_lock( &sound_mutex );
		if (!mz -> soundmode || mz -> playStop || mz -> pause
				|| (mz -> howipl > 0 && mz -> howipl != 10))
			tpause = TRUE;
		else
			tpause = FALSE;
		tclk = get_internal_clock();
		if (!tpause) {
			posll = (tclk - sound_baseclock) * (sound_datasize / sound_bytes)
				/ (get_frameclk_z80() * sound_multiply);
			pos = (int)posll * sound_bytes;
		} else
			pos = sound_datasize;
		if (pos > sound_bufsize)
			pos = sound_bufsize;
		if (pos <= 0) {
			if (!sound_thread_realtime)
				pthread_mutex_unlock( &sound_mutex );
			usleep( MZ2000_REFRESHRATE / 2 * 1000 );
			continue;
		}
		posdiff = 0;
		posnct = 0;
		n = 0;
		if (0 && mz2000_global -> enable_soundtminterpolations) {
			if (sound_stat_oldvalue1 > 0 && sound_stat_oldvalue2 > 0) {
				int tlen = pos;

#if 1	/* absolute */
				posdiff = sound_stat_oldvalue1 - sound_playoffset;
#else	/* relative */
				posdiff = sound_stat_oldvalue1 - sound_stat_oldvalue2;
#endif
				if (posdiff) {
#if 1	/* normals */
					n = (int)(tlen + 2) / abs(posdiff);
#else	/* max=1 */
					n = (int)(tlen + 2) / 2;
#endif
					if (abs(n < 3))
						n = 3;
					if (posdiff < 0)
						n = -n;
				}
			}
#if 0
			{
				static int ct = 0;

				ct++;
				if (0 || ct > 10) {
					_TRACE( "oval1=%d oval2=%d sbuf-ofs.=%d diff=%d n=%d lenBytes=%d\n",
					sound_stat_oldvalue1, sound_stat_oldvalue2, sound_playoffset,
					posdiff, n, pos );
					ct = 0;
				}
			}
#endif
		}
		if (sound_bytes == 1) {
			if (!tpause) {
				c = sound_playdata8;
				cp = sound_playbuf;
				for (i = 0, j = 0; i < pos; j++) {
					if (i < sound_bufsize) {
						if (sound_attrbuf[i])
							c = sound_databuf[i];
					} else
						c = SOUND_PULSE8_M;
					if (n && !(j % n) && i < pos - 2) {
						if (n >= 0) {
							*cp++ = c;
							*cp++ = c;
							posnct++;
							i += 2;
						} else
							--posnct;
					} else {
						*cp++ = c;
						i++;
					}
				}
				sound_playdata8 = c;
			} else {
				c = SOUND_PULSE8_M;
				cp = sound_playbuf;
				for (i = 0; i < pos; i++)
					*cp++ = c;
				sound_playdata8 = c;
			}
			pos += posnct;
			if (pos < sound_bufsize && !tpause) {
				int tt;

				memcpy( sound_databuf, sound_databuf + pos, sound_bufsize - pos );
				memcpy( sound_attrbuf, sound_attrbuf + pos, sound_bufsize - pos );
				tt = sound_bufsize - pos;
				memset( sound_databuf + tt, SOUND_PULSE8_M, sound_bufsize - tt );
				memset( sound_attrbuf + tt, 0, sound_bufsize - tt );
			} else {
				memset( sound_databuf, SOUND_PULSE8_M, sound_bufsize );
				memset( sound_attrbuf, 0, sound_bufsize );
			}
		} else {
			int tpos = pos / 2;

			if (!tpause) {
				c2 = sound_playdata16 & 0xff;		/* lower */
				c1 = (sound_playdata16 >> 8) & 0xff;	/* upper */
				cp = sound_playbuf;
				for (i = 0, j = 0; i < pos / 2; j+=2 /* sound bytes */) {
					if (i < sound_bufsize/2) {
						if (sound_attrbuf[i]) {
							c2 = sound_databuf[i * 2];
							c1 = sound_databuf[i * 2 + 1];
						}
					} else {
						c2 = SOUND_PULSE16_M;
						c1 = SOUND_PULSE16_M;
					}
					if (n && !(j % n) && i < (int)pos/2 - 2) {
						if (n >= 0) {
							*cp++ = c2;
							*cp++ = c1;
							*cp++ = c2;
							*cp++ = c1;
							posnct += 2;
							i += 2;
						} else
							posnct -= 2;
					} else {
						*cp++ = c2;
						*cp++ = c1;
						i++;
					}
				}
				sound_playdata16 = c2 | (c1 << 8);
			} else {
				c = SOUND_PULSE16_M;
				cp = sound_playbuf;
				for (i = 0; i < pos / 2; i++)
					*cp++ = c;
				sound_playdata16 = c;
			}
			pos += posnct;
			if (pos < sound_bufsize  && !tpause) {
				int tt, tt2;

				memcpy( sound_databuf, sound_databuf + pos, sound_bufsize - pos );
				memcpy( sound_attrbuf, sound_attrbuf + tpos, sound_bufsize - tpos );
				tt  = sound_bufsize - pos;
				tt2 = sound_bufsize / 2 - tpos;
				memset( sound_databuf + tt, SOUND_PULSE16_M, sound_bufsize - tt );
				memset( sound_attrbuf + tt2, 0, sound_bufsize - tt2 );
			} else {
				memset( sound_databuf, SOUND_PULSE16_M, sound_bufsize );
				memset( sound_attrbuf, 0, sound_bufsize / 2 );
			}
		}
		ridx = esd_playindex;
		if (ridx < sound_writeindex) {
			if (sound_writeindex - ridx < sound_playleftmin
					|| sound_writeindex - ridx > sound_playleftmax) {
				sound_writeindex = (ridx + sound_playoffset) % sound_playbuffer;
#ifdef ENABLE_DISPBUFFER
				fprintf( stderr, "sound-buffer: writeindex reset - A\n" );
#endif
			}
		} else {
			if (sound_writeindex + sound_playbuffer - ridx < sound_playleftmin
					|| sound_writeindex + sound_playbuffer - ridx > sound_playleftmax) {
				sound_writeindex = (ridx + sound_playoffset) % sound_playbuffer;
#ifdef ENABLE_DISPBUFFER
				fprintf( stderr, "sound-buffer: writeindex reset - B\n" );
#endif
			}
		}
		sound_stat_oldvalue2 = sound_stat_oldvalue1;
		if (esd_playindex < sound_writeindex)
			sound_stat_value1 = sound_writeindex - esd_playindex;
		else
			sound_stat_value1 = sound_writeindex + sound_playbuffer - esd_playindex;
		widx = sound_writeindex;
		if (widx + pos <= sound_playbuffer)
			memcpy( &esd_playbuf[widx], sound_playbuf, pos );
		else {
			int tt = sound_playbuffer - widx;

			memcpy( &esd_playbuf[widx], sound_playbuf, tt );
			memcpy( esd_playbuf, &sound_playbuf[tt], pos - tt );
		}
		sound_writeindex = (widx + pos) % sound_playbuffer;
		if (esd_playindex < sound_writeindex)
			sound_stat_value1 = sound_writeindex - esd_playindex;
		else
			sound_stat_value1 = sound_writeindex + sound_playbuffer - esd_playindex;
		sound_stat_oldvalue1 = sound_stat_value1;
		sound_baseclock = tclk;
		if (!sound_thread_realtime)
			pthread_mutex_unlock( &sound_mutex );
		usleep( MZ2000_REFRESHRATE / 2 * 1000 );
	}
	soundthread_waitstatus = FALSE;
	return;
}

static void sound_thread_2( MZ2000 *mz )
{
	int i, r, bleft;

	soundthread2_stopstatus = FALSE;
	while (!soundthread2_stopstatus) {
		if (!esd_available) {
			usleep( (mz -> sound_playbuffer_ms) * 1000 );
			continue;
		}
		if (!(mz -> soundmode) || mz -> playStop || mz -> pause) {
			sound_reset( mz );
			usleep( (mz -> sound_playbuffer_ms) * 1000 );
			continue;
		}
		esd_playindex = 0;
		bleft = sound_playbuffer;
#ifdef ENABLE_DISPBUFFER
		fprintf( stderr, "sound-buffer:%d<%d<%d (%d=%d/%d)\n",
			sound_stat_value2, sound_stat_value1, sound_stat_value3,
			sound_playbuffer/SOUND_PLAYBUFFER_DIV,
			sound_playbuffer, SOUND_PLAYBUFFER_DIV );
#endif
		for (i = 0; i < SOUND_PLAYBUFFER_DIV; i++) {
			if (bleft >= sound_playbuffer - SOUND_PLAYBUFFER_DIV) {
				r = write( esd_sock2, esd_playbuf + esd_playindex,
					sound_playbuffer/SOUND_PLAYBUFFER_DIV );
				if (r > 0)
					esd_playindex += r;
				else
					fprintf( stderr, "Warning: sound write error.\n" );
			} else {
				r = write( esd_sock2, esd_playbuf + esd_playindex,
					bleft );
				if (r > 0)
					esd_playindex = 0;
				else
					fprintf( stderr, "Warning: sound write error.\n" );
			}
#if 0
			while (!z80_timer()) {
				printf("-w");
				usleep( 1000 );
			}
#endif
			if (!esd_playindex || soundthread2_stopstatus)
				break;
		}
	}
	soundthread2_waitstatus = FALSE;
	return;
}

static int sound_startthread( MZ2000 *mz )
{
	int thread_policy;
	pthread_attr_t thread_attr;
	struct sched_param thread_param;

	sound_thread_realtime1 = FALSE;
	soundthread_waitstatus = TRUE;
	soundthread_stopstatus = TRUE;
#if 0
	if (pthread_attr_init( &thread_attr )) {
		fprintf( stderr, "sound_startthread_1 error (pthread_attr_init)\n");
		return FALSE;
	}
	if (pthread_attr_getschedpolicy( &thread_attr, &thread_policy ))
		fprintf( stderr, "sound_startthread_1 error (pthread_getschedpolicy)\n" );
	if (pthread_attr_getschedparam( &thread_attr, &thread_param ))
		fprintf( stderr, "sound_startthread_1 error (thread_attr_getschedparam)\n" );
	if (pthread_attr_setschedpolicy( &thread_attr, SCHED_FIFO )) {
		fprintf( stderr, "sound_startthread_1 error (pthread_setschedpolicy)\n" );
	} else {
		int t;

		if ((t = sched_get_priority_max( SCHED_FIFO )) == -1) {
			fprintf( stderr, "sound_startthread_1 error (sched_get_priority_max)\n" );
		} else {
			thread_param.sched_priority = t - 1;
			if (pthread_attr_setschedparam( &thread_attr, &thread_param )) {
				fprintf( stderr, "sound_startthread_1 error (pthread_setschedparam)\n" );
				return FALSE;
			}
		}
		if (pthread_attr_setinheritsched( &thread_attr, PTHREAD_EXPLICIT_SCHED )) {
			fprintf( stderr, "sound_startthread_1 error (pthread_attr_setinheritsched)\n" );
			return FALSE;
		}
	}
	if (!pthread_create( &soundthread_id, &thread_attr, (void *)sound_thread, (void *)mz )) {
		while (soundthread_stopstatus)
			usleep( 100 * 1000 );
		sound_thread_realtime1 = TRUE;
		return TRUE;
	}
	/* fprintf( stderr, "sound_startthread: Warning SCHED_FIFO thread failed, try SCHED_OTHER ...\n"); */
#endif
	if (pthread_attr_init( &thread_attr )) {
		fprintf( stderr, "sound_startthread error (pthread_attr_init)\n");
		return FALSE;
	}
	if (pthread_attr_getschedpolicy( &thread_attr, &thread_policy ))
		fprintf( stderr, "sound_startthread_1 error (pthread_getschedpolicy)\n" );
	if (pthread_attr_getschedparam( &thread_attr, &thread_param ))
		fprintf( stderr, "sound_startthread_1 error (thread_attr_getschedparam)\n" );
	if (pthread_attr_setschedpolicy( &thread_attr, SCHED_OTHER )) {
		fprintf( stderr, "sound_startthread_1 error (pthread_setschedpolicy)\n" );
	} else {
		int t;

		if ((t = sched_get_priority_max( SCHED_OTHER )) == -1) {
			fprintf( stderr, "sound_startthread_1 error (sched_get_priority_max)\n" );
		} else {
			thread_param.sched_priority = t;
			if (pthread_attr_setschedparam( &thread_attr, &thread_param )) {
				fprintf( stderr, "sound_startthread_1 error (pthread_setschedparam)\n" );
				return FALSE;
			}
		}
		if (pthread_attr_setinheritsched( &thread_attr, PTHREAD_EXPLICIT_SCHED )) {
			fprintf( stderr, "sound_startthread_1 error (pthread_attr_setinheritsched)\n" );
			return FALSE;
		}
	}
	if (pthread_create( &soundthread_id, &thread_attr, (void *)sound_thread, (void *)mz )) {
		fprintf( stderr, "sound_startthread_1 error (pthread_create)\n" );
		return FALSE;
	}
	while (soundthread_stopstatus)
		usleep( 100 * 1000 );
	return TRUE;
}

static int sound_startthread_2( MZ2000 *mz )
{
	int thread_policy;
	pthread_attr_t thread_attr;
	struct sched_param thread_param;

	sound_thread_realtime2 = FALSE;
	soundthread2_waitstatus = TRUE;
	soundthread2_stopstatus = TRUE;
#if 0
	if (pthread_attr_init( &thread_attr )) {
		fprintf( stderr, "sound_startthread_2 error (pthread_attr_init)\n");
		return FALSE;
	}
	if (pthread_attr_getschedpolicy( &thread_attr, &thread_policy ))
		fprintf( stderr, "sound_startthread_2 error (pthread_getschedpolicy)\n" );
	if (pthread_attr_getschedparam( &thread_attr, &thread_param ))
		fprintf( stderr, "sound_startthread_2 error (thread_attr_getschedparam)\n" );
	if (pthread_attr_setschedpolicy( &thread_attr, SCHED_FIFO )) {
		fprintf( stderr, "sound_startthread_2 error (pthread_setschedpolicy)\n" );
	} else {
		int t;

		if ((t = sched_get_priority_max( SCHED_FIFO )) == -1) {
			fprintf( stderr, "sound_startthread_2 error (sched_get_priority_max)\n" );
		} else {
			thread_param.sched_priority = t - 1;
			if (pthread_attr_setschedparam( &thread_attr, &thread_param )) {
				fprintf( stderr, "sound_startthread_2 error (pthread_setschedparam)\n" );
				return FALSE;
			}
		}
		if (pthread_attr_setinheritsched( &thread_attr, PTHREAD_EXPLICIT_SCHED )) {
			fprintf( stderr, "sound_startthread_2 error (pthread_attr_setinheritsched)\n" );
			return FALSE;
		}
	}
	if (!pthread_create( &soundthread2_id, &thread_attr, (void *)sound_thread_2, (void *)mz )) {
		while (soundthread2_stopstatus)
			usleep( 100 * 1000 );
		sound_thread_realtime1 = TRUE;
		return TRUE;
	}
	/* fprintf( stderr, "sound_startthread_2: Warning SCHED_FIFO thread failed, try SCHED_OTHER ...\n"); */
#endif
	if (pthread_attr_init( &thread_attr )) {
		fprintf( stderr, "sound_startthread error (pthread_attr_init)\n");
		return FALSE;
	}
	if (pthread_attr_getschedpolicy( &thread_attr, &thread_policy ))
		fprintf( stderr, "sound_startthread_2 error (pthread_getschedpolicy)\n" );
	if (pthread_attr_getschedparam( &thread_attr, &thread_param ))
		fprintf( stderr, "sound_startthread_2 error (thread_attr_getschedparam)\n" );
	if (pthread_attr_setschedpolicy( &thread_attr, SCHED_OTHER )) {
		fprintf( stderr, "sound_startthread_2 error (pthread_setschedpolicy)\n" );
	} else {
		int t;

		if ((t = sched_get_priority_max( SCHED_OTHER )) == -1) {
			fprintf( stderr, "sound_startthread_2 error (sched_get_priority_max)\n" );
		} else {
			thread_param.sched_priority = t;
			if (pthread_attr_setschedparam( &thread_attr, &thread_param )) {
				fprintf( stderr, "sound_startthread_2 error (pthread_setschedparam)\n" );
				return FALSE;
			}
		}
		if (pthread_attr_setinheritsched( &thread_attr, PTHREAD_EXPLICIT_SCHED )) {
			fprintf( stderr, "sound_startthread_2 error (pthread_attr_setinheritsched)\n" );
			return FALSE;
		}
	}
	if (pthread_create( &soundthread2_id, &thread_attr, (void *)sound_thread_2, (void *)mz )) {
		fprintf( stderr, "sound_startthread_2 error (pthread_create)\n" );
		return FALSE;
	}
	while (soundthread2_stopstatus)
		usleep( 100 * 1000 );
	return TRUE;
}

int sound_playwav_file( MZ2000 *mz, SOUND_WAVLIST no )
{
	if (no < 0 || no >= sizeof(sound_wavedata)/sizeof(SOUND_WAVEDATA))
		return -1;
	if (sound_wavedata[no].id < 0)
		return -1;
	if (!(mz -> soundmode))
		return 0;
	if (sound_wavedata[no].device_type == 1 && !(mz -> enable_soundtape))
		return 0;
	if (sound_wavedata[no].device_type == 2 && !(mz -> enable_soundfdd))
		return 0;
	esd_sample_play( esd_sock, sound_wavedata[no].id );
	return 0;
}

static char *sound_loadwav_file( WAVEDATA_STRUCT *ws_p, const char *fn )
{
	int r;
	unsigned long len;
	char *p, *rp;

	r = wavefile_ropen( WAVEFILE_ATTR_DEFAULT, ws_p, fn );
	if (r < 0) {
#ifdef EMZ2000_LIBDIR
		char tpath[PATH_MAX];

		strncpy( tpath, EMZ2000_LIBDIR "/", PATH_MAX );
		tpath[PATH_MAX - 1] = '\0';
		strncat( tpath, fn, PATH_MAX - strlen(tpath) - 1 );
		r = wavefile_ropen( WAVEFILE_ATTR_DEFAULT, ws_p, tpath );
		if (r < 0) {
#endif
			fprintf( stderr, "%s: read open error (result=%d) \"%s\"\n", module_name, r, fn );
			return NULL;
#ifdef EMZ2000_LIBDIR
		}
#endif
	}
	len = wavedata_getlenbytes( ws_p );
	p = MALLOC( len );
	if (!p) {
		fprintf( stderr, "%s: memory allocation error (%lu bytes).\n", module_name, len );
		wavefile_rclose( ws_p );
		return NULL;
	}
	rp = p;
	while (len > 0) {
		r = read( wavedata_getfd(ws_p), p, len );
		if (r > 0) {
			len -= r;
			p += r;
		}
		if (r < 0) {
			fprintf( stderr, "%s: read error r=%d (%s)\n", module_name, r, strerror(errno) );
			wavefile_rclose( ws_p );
			return NULL;
		}
	}
	r = wavefile_rclose( ws_p );
	if (r < 0) {
		fprintf( stderr, "%s: read close error\n", module_name );
		return NULL;
	}
	return rp;
}

static void sound_freewav( void )
{
	int i;

	for (i = 0; i < sizeof(sound_wavedata)/sizeof(SOUND_WAVEDATA); i++) {
		if (sound_wavedata[i].id >= 0) {
			esd_sample_free( esd_sock, sound_wavedata[i].id );
			sound_wavedata[i].id = -1;
		}
	}
	return;
}

static int sound_loadwav( MZ2000 *mz )
{
	int r, i, len, id, trate;
	char *p, esdname[ESD_NAME_MAX];
	WAVEDATA_STRUCT ws;
	esd_format_t tformat;

	trate = sound_samplingrate;
	for (i = 0; i < sizeof(sound_wavedata)/sizeof(SOUND_WAVEDATA); i++) {
		_TRACE("%d: %s\n", i, sound_wavedata[i].fn);
		wavedata_initstruct( &ws );
		p = sound_loadwav_file( &ws, sound_wavedata[i].fn );
		if (!p)
			continue;
		_TRACE("f=%d ch=%d bytes=%d\n", ws.f, ws.ch, ws.bytes );
		tformat = ESD_STREAM | ESD_PLAY;
		if (ws.ch == 1)
			tformat |= ESD_MONO;
		else
			tformat |= ESD_STEREO;
		if (ws.bytes == 1)
			tformat |= ESD_BITS8;
		else
			tformat |= ESD_BITS16;
		trate = ws.f;
		len = wavedata_getlenbytes( &ws );
		strncpy( esdname, sound_wavedata[i].fn, ESD_NAME_MAX );
		esdname[ESD_NAME_MAX - 1] = '\0';
		id = esd_sample_cache( esd_sock, tformat, trate, len, esdname );
		if (id >= 0) {
			r = write( esd_sock, p, len );
			if (r < 0)
				fprintf( stderr, "The WAV file \"%s\" reading failed.\n", esdname );
			esd_confirm_sample_cache( esd_sock );
			esd_set_stream_pan( esd_sock, id, sound_volume2, sound_volume2 );
			sound_wavedata[i].id = id;
			_TRACE("ok:id=%d\n", id);
		} else
			FREE( p );
	}
	return 0;
}

int sound_enablechk( void )
{
	return esd_available;
}

int sound_updateparams( MZ2000 *mz )
{
	sound_exit();
	return sound_init( mz );
}

void sound_exit( void )
{
	int i;

	if (!esd_available) {
		if (esd_playbuf) {
			FREE( esd_playbuf );
			esd_playbuf = NULL;
		}
		if (sound_databuf) {
			FREE( sound_databuf );
			sound_databuf = NULL;
		}
		if (sound_attrbuf) {
			FREE( sound_attrbuf );
			sound_attrbuf = NULL;
		}
		if (sound_playbuf) {
			FREE( sound_playbuf );
			sound_playbuf = NULL;
		}
		return;
	}
	soundthread_stopstatus = TRUE;
	for (i = 0; i < 20 && soundthread_waitstatus; i++)
		usleep( 100 * 1000 );
	soundthread2_stopstatus = TRUE;
	for (i = 0; i < 20 && soundthread2_waitstatus; i++)
		usleep( 100 * 1000 );
	if (esd_sock2 >= 0) {
		close(esd_sock2);
		esd_sock2 = -1;
	}
	sound_freewav();
	if (esd_ctrl) {
		esd_free_all_info( esd_ctrl );
		esd_ctrl = NULL;
	}
	if (esd_sock >= 0) {
		close(esd_sock);
		esd_sock = -1;
	}
	if (esd_playbuf) {
		FREE( esd_playbuf );
		esd_playbuf = NULL;
	}
	if (sound_databuf) {
		FREE( sound_databuf );
		sound_databuf = NULL;
	}
	if (sound_attrbuf) {
		FREE( sound_attrbuf );
		sound_attrbuf = NULL;
	}
	if (sound_playbuf) {
		FREE( sound_playbuf );
		sound_playbuf = NULL;
	}
	esd_available = FALSE;
	return;
}

int sound_init( MZ2000 *mz )
{
	unsigned int t;

	sound_volume1 = mz -> sound_volume1;	/* MZ pulse sound */
	sound_volume2 = mz -> sound_volume2;	/* wave sound effects */
	t = (mz -> sound_playbuffer_ms + MZ2000_REFRESHRATE - 1) / MZ2000_REFRESHRATE;
	t = t * MZ2000_REFRESHRATE;
	sound_samplingrate = mz -> sound_freq;
	sound_bytes        = mz -> sound_bit / 8;
	sound_multiply  = t / MZ2000_REFRESHRATE;
	sound_delaytime = mz -> sound_delaytime_ms;

	sound_datasize = (sound_samplingrate * sound_multiply * MZ2000_REFRESHRATE)
		/ 1000 * sound_bytes;
	sound_bufsize = (sound_datasize + 512);
	t = (sound_samplingrate * sound_delaytime / 1000) * sound_bytes;
#if 1	/* New Buffer Values */
	sound_playbuffer  = sound_datasize + t;
	sound_playoffset  = sound_playbuffer - t / 2;
	sound_playleftmax = sound_playbuffer - t / 20;
	sound_playleftmin = t / 20;
#else	/* OLD Calcuration Algorithms (use for debug) */
	t = sound_datasize;
	sound_playbuffer = t * 4;
	sound_playoffset = t * 2 - t / 4;
	sound_playleftmax = t * 3 + t / 4 * 3;
	sound_playleftmin = t / 2;
#endif
	sound_stat_value1 = 0;
	sound_stat_value2 = sound_playleftmin;
	sound_stat_value3 = sound_playleftmax;

	pthread_mutex_init( &sound_mutex, NULL );
	if (sound_bytes == 1)
		esd_format = ESD_BITS8  | ESD_MONO | ESD_STREAM | ESD_PLAY;
	else
		esd_format = ESD_BITS16 | ESD_MONO | ESD_STREAM | ESD_PLAY;
	esd_rate = sound_samplingrate;
	esd_sock = esd_open_sound( NULL /* host */ );
	if (esd_sock < 0) {
		esd_available = FALSE;
		return FALSE;
	}
	sound_databuf  = MALLOC( sound_bufsize );
	if (!sound_databuf) {
		close(esd_sock);
		esd_sock = -1;
		esd_available = FALSE;
		return FALSE;
	}
	memset( sound_databuf, sound_bytes == 1 ? SOUND_PULSE8_M : SOUND_PULSE16_M, sound_bufsize );
	sound_attrbuf = MALLOC( sound_bufsize );
	if (!sound_attrbuf) {
		FREE( sound_databuf );
		sound_databuf = NULL;
		close( esd_sock );
		esd_sock = -1;
		esd_available = FALSE;
		return FALSE;
	}
	memset( sound_attrbuf, 0, sound_bufsize );
	sound_playbuf = MALLOC( sound_bufsize );
	if (!sound_playbuf) {
		FREE( sound_attrbuf );
		sound_attrbuf = NULL;
		FREE( sound_databuf );
		sound_databuf = NULL;
		close(esd_sock);
		esd_sock = -1;
		esd_available = FALSE;
		return FALSE;
	}
	memset( sound_playbuf, sound_bytes == 1 ? SOUND_PULSE8_M : SOUND_PULSE16_M, sound_bufsize );
	esd_playbuf = MALLOC( sound_playbuffer );
	if (!esd_playbuf) {
		FREE( sound_playbuf );
		sound_playbuf = NULL;
		FREE( sound_attrbuf );
		sound_attrbuf = NULL;
		FREE( sound_databuf );
		sound_databuf = NULL;
		close(esd_sock);
		esd_sock = -1;
		esd_available = FALSE;
		return FALSE;
	}
	memset( esd_playbuf, sound_bytes == 1 ? SOUND_PULSE8_M : SOUND_PULSE16_M, sound_playbuffer );
	esd_sock2 = esd_play_stream_fallback( esd_format, esd_rate,
		NULL, "MZ-2000_sound" );
	if (esd_sock2 <= 0) {
		FREE( esd_playbuf );
		esd_playbuf = NULL;
		FREE( sound_playbuf );
		sound_playbuf = NULL;
		FREE( sound_attrbuf );
		sound_attrbuf = NULL;
		FREE( sound_databuf );
		sound_databuf = NULL;
		close(esd_sock);
		esd_sock = -1;
		esd_available = FALSE;
		return FALSE;
	}
	esd_ctrl = esd_get_all_info( esd_sock );
	if (esd_ctrl) {
		if (esd_ctrl -> player_list)
			esd_set_stream_pan( esd_sock, esd_ctrl -> player_list -> source_id,
				sound_volume1, sound_volume1 );
		else
			esd_set_stream_pan( esd_sock, 0,
				sound_volume1, sound_volume1 );
		/* esd_print_all_info( esd_ctrl ); */
		esd_free_all_info( esd_ctrl );
		esd_ctrl = NULL;
	}
	esd_playindex = 0;
	sound_thread_realtime = FALSE;
	if (!sound_startthread( mz )) {
		FREE( esd_playbuf );
		esd_playbuf = NULL;
		FREE( sound_playbuf );
		sound_playbuf = NULL;
		FREE( sound_attrbuf );
		sound_attrbuf = NULL;
		FREE( sound_databuf );
		sound_databuf = NULL;
		close(esd_sock2);
		esd_sock2 = -1;
		close(esd_sock);
		esd_sock = -1;
		esd_available = FALSE;
		return FALSE;
	}
	if (!sound_startthread_2( mz )) {
		FREE( esd_playbuf );
		esd_playbuf = NULL;
		FREE( sound_playbuf );
		sound_playbuf = NULL;
		FREE( sound_attrbuf );
		sound_attrbuf = NULL;
		FREE( sound_databuf );
		sound_databuf = NULL;
		close(esd_sock2);
		esd_sock2 = -1;
		close(esd_sock);
		esd_sock = -1;
		esd_available = FALSE;
		return FALSE;
	}
	if (sound_thread_realtime1 && sound_thread_realtime2)
		sound_thread_realtime = TRUE;
	esd_available = TRUE;
	sound_loadwav( mz );
	return TRUE;
}

/*
	Local Variables:
	mode:c++
	c-set-style:"k&r"
	c-basic-offset:8
	tab-width:8
	End:
*/
