/* File: state.c

	SHARP MZ-2000/2200/80B/B2 Emulator "emz2000"
		States Save/Load Module

	Copyright (C) 2002-2019 FUKUI, Toshio.
	This file is part of the emz2000 software.
	See copyright notice in the COPYING file.
*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <unistd.h>

#include "common.h"
#include "mz2000.h"
#include "uigtk.h"
#include "plugin.h"

static int save_state_main( MZ2000 *mz, const char *filename )
{
	int i, r;
	FILE *fp;
	unsigned char head[16], info[16+PLUGIN_GUIDLENGTH_MAXNUM];
	unsigned int val;

	fp = fopen( filename, "w" );
	if (!fp) {
		uigtk_messagebox( "Save states error. (fopen)",
			"States", UIGTK_BUTTON_OKONLY | UIGTK_ICON_ERROR );
		return -1;
	}
	memset( head, 0, sizeof(head) );
	head[0] = 'E';
	head[1] = 'M';
	head[2] = 'U';
	head[3] = 'Z';
	head[4] = 2;	/* Type:    0=all, 1=Windows only, 2=Linux/UNIX only */
	head[5] = 0;	/* reserved */
	head[6] = 2;	/* Sub Version: 0<=0.8f, 1>=0.8g, 2>=0.92 */
	head[7] = 0;	/* Version:     0=first */
	r = fwrite( head, 1, sizeof(head), fp );
	if (r < sizeof(head)) {
		fclose( fp );
		return -2;
	}
	/* structure Z80 */
	val = sizeof( z80 );
	r = fwrite( (void *)&val, 1, sizeof(val), fp );
	if (r < sizeof(val)) {
		fclose( fp );
		return -2;
	}
	r = fwrite( (void *)&z80, 1, sizeof(Z80), fp );
	if (r < sizeof(Z80)) {
		fclose( fp );
		return -2;
	}
	/* structure MZ */
	val = sizeof( MZ2000 );
	r = fwrite( (void *)&val, 1, sizeof(val), fp );
	if (r < sizeof(val)) {
		fclose( fp );
		return -2;
	}
	r = fwrite( mz, 1, sizeof(MZ2000), fp );
	if (r < sizeof(MZ2000)) {
		fclose( fp );
		return -2;
	}
	/* structure MZIOInfo */
	val = sizeof( MZIOInfo );
	r = fwrite( (void *)&val, 1, sizeof(val), fp );
	if (r < sizeof(val)) {
		fclose( fp );
		return -2;
	}
	r = fwrite( mz->ioinfo, 1, sizeof(MZIOInfo), fp );
	if (r < sizeof(MZIOInfo)) {
		fclose( fp );
		return -2;
	}
	/* structure MZScreen */
	val = sizeof( MZScreen );
	r = fwrite( (void *)&val, 1, sizeof(val), fp );
	if (r < sizeof(val)) {
		fclose( fp );
		return -2;
	}
	r = fwrite( mz->screen, 1, sizeof(MZScreen), fp );
	if (r < sizeof(MZScreen)) {
		fclose( fp );
		return -2;
	}
	r = fwrite( mz->memory, 1, MZ_MAINMEMORY_SIZE, fp );
	if (r < MZ_MAINMEMORY_SIZE) {
		fclose( fp );
		return -2;
	}
	r = fwrite( mz->vram, 1, MZ_TEXTVRAM_SIZE, fp );
	if (r < MZ_TEXTVRAM_SIZE) {
		fclose( fp );
		return -2;
	}
	r = fwrite( mz->gram1, 1, MZ_GRAMPAGE_SIZE, fp );
	if (r < MZ_GRAMPAGE_SIZE) {
		fclose( fp );
		return -2;
	}
	r = fwrite( mz->gram2, 1, MZ_GRAMPAGE_SIZE, fp );
	if (r < MZ_GRAMPAGE_SIZE) {
		fclose( fp );
		return -2;
	}
	r = fwrite( mz->gram3, 1, MZ_GRAMPAGE_SIZE, fp );
	if (r < MZ_GRAMPAGE_SIZE) {
		fclose( fp );
		return -2;
	}

	/* chunks for DLL work area */
	memset( &info, 0, sizeof(info) );
	info[0] = 'I';
	info[1] = 'N';
	info[2] = 'F';
	info[3] = 'O';
	for (i = 0; i < PLUGIN_BOARD_MAXNUM; i++) {
		unsigned int size;
		int ver;
		char *buf, *ident;

		ident = NULL;
		buf = NULL;
		if (plugin_get_boardstateinfo( i, &ident, &ver, &size, (unsigned char **)&buf ) && ident && size > 0 && buf) {
			*((int *)(&info[4]))   = ver;	/* ver. */
			*((int *)(&info[12]))  = 0;
			*((unsigned int *)(&info[8])) = size;	/* size */
			strncpy( (char *)&info[16], ident, PLUGIN_GUIDLENGTH_MAXNUM );
			info[16 + PLUGIN_GUIDLENGTH_MAXNUM - 1] = '\0';
			r = fwrite( info, 1, sizeof(info), fp );
			if (r < sizeof(info)) {
				fclose( fp );
				return -2;
			}
			r = fwrite( buf, 1, size, fp );
			if (r < size) {
				fclose( fp );
				return -2;
			}
		}
	}
	*((int *)(&info[4])) = 0;	/* ver. */
	*((int *)(&info[12]))  = 0;
	*((unsigned int *)(&info[8])) = 0;	/* size (DATA END) */
	r = fwrite( info, 1, 16, fp );
	if (r < 16) {
		fclose( fp );
		return -2;
	}
	fclose( fp );
	return 0;
}

static int load_state_main( MZ2000 *mz, const char *filename )
{
	int r;
	char *buf;
	FILE *fp;
	Z80 tz80;
	MZ2000 tmz;
	MZScreen tscreen;
	unsigned char head[16], thead[16];
	unsigned char info[16], tinfo[16];
	unsigned int val;
	unsigned int ver, subver;

	fp = fopen( filename, "r" );
	if (!fp) {
		uigtk_messagebox( "Load states error. (fopen)",
			"States", UIGTK_BUTTON_OKONLY | UIGTK_ICON_ERROR );
		return -1;
	}
	memset( head, 0, sizeof(head) );
	head[0] = 'E';
	head[1] = 'M';
	head[2] = 'U';
	head[3] = 'Z';
	r = fread( thead, 1, sizeof(head), fp );
	if (r < sizeof(head) || strncmp( (char *)thead, (char *)head, 4 )) {
		fclose( fp );
		uigtk_messagebox( "Load states error. (Eyecache Error)",
			"States", UIGTK_BUTTON_OKONLY | UIGTK_ICON_ERROR );
		return -1;
	}
	if (thead[5] || !thead[4]) {		/* Type == 0 */
		fclose( fp );
		uigtk_messagebox( "Load states error. (Architecture Error, common type not supported)",
			"States", UIGTK_BUTTON_OKONLY | UIGTK_ICON_ERROR );
		return -1;
	}
	if (thead[5] || thead[4] != 2) {	/* type <>2 */
		fclose( fp );
		uigtk_messagebox( "Load states error. (Architecture Error, not Linux/UNIX)",
			"States", UIGTK_BUTTON_OKONLY | UIGTK_ICON_ERROR );
		return -1;
	}
	ver    = thead[7];
	subver = thead[6];
	if (ver != 0 || subver >= 3)	 {	/* >= version 0.3 */
		fclose( fp );
		uigtk_messagebox( "Load states error. (Unsupported version, >=0.3)",
			"States", UIGTK_BUTTON_OKONLY | UIGTK_ICON_ERROR );
		return -1;
	}
	/* structure Z80 */
	r = fread( (void *)&val, 1, sizeof(val), fp );
	if (r < sizeof(val)) {
		fclose( fp );
		return -2;
	}
	if (val != sizeof(z80)) {
		fclose( fp );
		_TRACE("Z80 structure error : %d <> %d\n", val, sizeof(z80));
		return -3;
	}
	r = fread( (void *)&tz80, 1, sizeof(Z80), fp );
	if (r < sizeof(Z80)) {
		fclose( fp );
		return -2;
	}
	/* structure MZ */
	r = fread( (void *)&val, 1, sizeof(val), fp );
	if (r < sizeof(val)) {
		fclose( fp );
		return -2;
	}
	if (val != sizeof(MZ2000)) {
		fclose( fp );
		_TRACE("MZ2000 structure error : %d <> %d\n", val, sizeof(z80));
		return -3;
	}
	r = fread( &tmz, 1, sizeof(MZ2000), fp );
	if (r < sizeof(MZ2000)) {
		fclose( fp );
		return -2;
	}
	/* structure MZIOInfo - I */
	r = fread( (void *)&val, 1, sizeof(val), fp );
	if (r < sizeof(val)) {
		fclose( fp );
		return -2;
	}
	if (val != sizeof(MZIOInfo)) {
		fclose( fp );
		_TRACE("MZIOInfo structure error : %d <> %d\n", val, sizeof(z80));
		return -3;
	}
	/* update the system structures */
	mz -> mzmode = tmz . mzmode;
	reset_mz2000_sub( mz );
	reinit_io( mz );
	memcpy( (void *)&z80.a, &tz80.a, (unsigned long)&tz80.left_clock - (unsigned long)&tz80.a );
	memcpy( &mz->inipl, &tmz.inipl, (unsigned long)&tmz.playStop - (unsigned long)&tmz.inipl );
	memcpy( (void *)&mz->cpu_peripherals_start, (void *)&tmz.cpu_peripherals_start,
		(unsigned long)&tmz.cpu_peripherals_end - (unsigned long)&tmz.cpu_peripherals_start );
	if (mz -> keyint_func)
		mz -> keyint_func = mz2io_keyintfunc;
	if (!ver && !subver) {
		mz -> timer8253_count[0] = 0x7a12;
		mz -> timer8253_count[1] = 0xa8c0;
		mz -> timer8253_count[2] = 0x0000;
		mz -> timer8253_fast_mode = 0;
	}
	if (!ver && subver <= 1) {
		/* initialize the CPU peripherals */
		mz -> cpu_peripherals_start = CPUPERIPHERALS_INIT;
	}
	/* structure MZIOInfo - II */
	r = fread( mz->ioinfo, 1, sizeof(MZIOInfo), fp );
	if (r < sizeof(MZIOInfo)) {
		fclose( fp );
		mz -> needIpl = TRUE;
		stop_z80();
		return -2;
	}
	/* structure MZScreen - I */
	r = fread( (void *)&val, 1, sizeof(val), fp );
	if (r < sizeof(val)) {
		fclose( fp );
		mz -> needIpl = TRUE;
		stop_z80();
		return -2;
	}
	if (val != sizeof(MZScreen)) {
		fclose( fp );
		mz -> needIpl = TRUE;
		stop_z80();
		_TRACE("MZScreen structure error : %d <> %d\n", val, sizeof(z80));
		return -3;
	}
	r = fread( &tscreen, 1, sizeof(MZScreen), fp );
	if (r < sizeof(MZScreen)) {
		fclose( fp );
		mz -> needIpl = TRUE;
		stop_z80();
		return -2;
	}
	/* update the system structures II */
	memcpy( &(mz->screen->scrbuf), &tscreen.scrbuf, ((unsigned long)&tscreen.font - (unsigned long)&tscreen.scrbuf) );
	mz -> displayMode = !(mz -> screen -> isgreen);
	refresh_io( mz );
	r = fread( mz->memory, 1, MZ_MAINMEMORY_SIZE, fp );
	if (r < MZ_MAINMEMORY_SIZE) {
		fclose( fp );
		mz -> needIpl = TRUE;
		stop_z80();
		return -2;
	}
	r = fread( mz->vram, 1, MZ_TEXTVRAM_SIZE, fp );
	if (r < MZ_TEXTVRAM_SIZE) {
		fclose( fp );
		mz -> needIpl = TRUE;
		stop_z80();
		return -2;
	}
	r = fread( mz->gram1, 1, MZ_GRAMPAGE_SIZE, fp );
	if (r < MZ_GRAMPAGE_SIZE) {
		fclose( fp );
		mz -> needIpl = TRUE;
		stop_z80();
		return -2;
	}
	r = fread( mz->gram2, 1, MZ_GRAMPAGE_SIZE, fp );
	if (r < MZ_GRAMPAGE_SIZE) {
		fclose( fp );
		mz -> needIpl = TRUE;
		stop_z80();
		return -2;
	}
	r = fread( mz->gram3, 1, MZ_GRAMPAGE_SIZE, fp );
	if (r < MZ_GRAMPAGE_SIZE) {
		fclose( fp );
		mz -> needIpl = TRUE;
		stop_z80();
		return -2;
	}

	/* chunks for DLL work regions */
	memset( &info, 0, sizeof(info) );
	info[0] = 'I';
	info[1] = 'N';
	info[2] = 'F';
	info[3] = 'O';
	while (1) {
		unsigned int size;
		int ver;
		char ident[PLUGIN_GUIDLENGTH_MAXNUM];

		r = fread( tinfo, 1, 16, fp );
		if (r < 16 || strncmp( (char *)tinfo, (char *)info, 4 )) {
			fclose( fp );
			mz -> needIpl = TRUE;
			stop_z80();
			return -2;
		}
		ver  = *((int *)(&tinfo[4]));
		size = *((unsigned int *)(&info[8]));
		if (!size)
			break;
		r = fread( ident, 1, PLUGIN_GUIDLENGTH_MAXNUM, fp );
		if (r < PLUGIN_GUIDLENGTH_MAXNUM) {
			fclose( fp );
			mz -> needIpl = TRUE;
			stop_z80();
			return -2;
		}
		ident[PLUGIN_GUIDLENGTH_MAXNUM - 1] = '\0';
		buf = MALLOC( size );
		if (!buf) {
			fclose( fp );
			mz -> needIpl = TRUE;
			stop_z80();
			return -4;
		}
		r = fread( buf, 1, size, fp );
		if (r < size) {
			fclose( fp );
			mz -> needIpl = TRUE;
			stop_z80();
			return -2;
		}
		plugin_set_boardstateinfo( ident, ver, size, (unsigned char *)buf );
		FREE( buf );
	}
	fclose( fp );
	screen_sethighreso( mz, mz -> screen_highreso );
	return 0;
}

int save_state( MZ2000 *mz )
{
	int r;

	if (mz -> howipl > 0) {
		uigtk_messagebox( "Save states error. (Running IPL)",
			"States", UIGTK_BUTTON_OKONLY | UIGTK_ICON_ERROR );
		return -1;
	}
	if (!(mz -> sfilename_havewrite)) {
		uigtk_messagebox( "Save states error. (No filenames)",
			"States", UIGTK_BUTTON_OKONLY | UIGTK_ICON_ERROR );
		return -1;
	}
	r = save_state_main( mz, mz -> sfilename );
	switch (r) {
	case 0 :	/* normal ends */
		break;
	case -1 :	/* already MessageBox was displayed */
		return -1;
	case -2 :
		uigtk_messagebox( "Save states error. (fwrite)",
			"States", UIGTK_BUTTON_OKONLY | UIGTK_ICON_ERROR );
		return -1;
	default :
		if (r > 0)
			break;
		uigtk_messagebox( "Save states error. (Unknown Error)",
			"States", UIGTK_BUTTON_OKONLY | UIGTK_ICON_ERROR );
		return -1;
	}
	return 0;
}

int load_state( MZ2000 *mz )
{
	int r;

	mz -> sfilename_havewrite = FALSE;
	r = load_state_main( mz, mz -> sfilename );
	switch (r) {
	case 0 :	/* normal ends */
		break;
	case -1 :	/* Already MessageBox was displayed */
		return -1;
	case -2 :
		uigtk_messagebox( "Load states error. (fread)",
			"States", UIGTK_BUTTON_OKONLY | UIGTK_ICON_ERROR );
		return -1;
	case -3 :
		uigtk_messagebox( "Load states error. (Structure Size Error)",
			"States", UIGTK_BUTTON_OKONLY | UIGTK_ICON_ERROR );
		return -1;
	case -4 :
		uigtk_messagebox( "Load states error. (Chunk Malloc)",
			"States", UIGTK_BUTTON_OKONLY | UIGTK_ICON_ERROR );
		return -1;
	default :
		if (r > 0)
			break;
		uigtk_messagebox( "Load states error. (Unknown Error)",
			"States", UIGTK_BUTTON_OKONLY | UIGTK_ICON_ERROR );
		return -1;
	}
	return 0;
}

/*
	Local Variables:
	mode:c++
	c-set-style:"k&r"
	c-basic-offset:8
	tab-width:8
	End:
*/
