/* File: tuchar.c

	Code Converter and Other Functions for Unicode/UCS2or4
		(require libiconv)

	Copyright (C) 2000-2019 FUKUI, Toshio.
	This file is part of the emz2000 software.
	See copyright notice in the COPYING file.
*/

#include "config.h"

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#include <fcntl.h>
#ifdef HAVE_STRING_H
#include <string.h>
#else
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif
#endif /* HAVE_STRING_H */
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_ICONV_H
#include <iconv.h>
#else
#ifdef HAVE_WCHAR_H
#include <wchar.h>
#endif
#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif
typedef void * iconv_t;
#endif

#ifndef HAVE_STRING_H
#define strrchr rindex
#endif
#ifdef WIN32
#define strcasecmp _stricmp
#endif

#include "common.h"
#include "tuchar.h"

#define TUCHAR_MALLOC(_X, _Y)		MALLOC((_X))
#define TUCHAR_REALLOC(_P, _X, _Y)	REALLOC((_P), (_X))
#define TUCHAR_FREE(_X, _Y)		FREE((_X))

#ifdef TUCHAR_USE_UCS4
#define UCSNAME "UCS-4LE"
#else
#define UCSNAME "UCS-2LE"
#endif
#define UTF8NAME "utf8"

#ifdef DEBUG
#undef TUCHAR_DISP_TERMINALMODE
#endif

static iconv_t id_l2u = NULL;
static iconv_t id_u2l = NULL;
static iconv_t id_f2u = NULL;
static iconv_t id_u2f = NULL;
static iconv_t id_utf2u = NULL;
static iconv_t id_u2utf = NULL;
static char term_codename_default[LNMAX];
static char term_codename_buf[LNMAX];
static char file_codename_buf[LNMAX];
static const char *term_codename;
static const char *file_codename;

int tuchar_isdigit( TUCHAR ch )
{
	if (ch >= 0x7f)
		return FALSE;
	return isdigit(ch);
}

int tuchar_isalpha( TUCHAR ch )
{
	if (ch >= 0x7f)
		return FALSE;
	return isalpha(ch);
}

int tuchar_isalnum( TUCHAR ch )
{
	if (ch >= 0x7f)
		return FALSE;
	return isalnum(ch);
}

int tuchar_islower( TUCHAR ch )
{
	if (ch >= 0x7f)
		return FALSE;
	return islower(ch);
}

int tuchar_isupper( TUCHAR ch )
{
	if (ch >= 0x7f)
		return FALSE;
	return isupper(ch);
}

int tuchar_isgraph( TUCHAR ch )
{
	if (ch >= 0x80)
		return TRUE;
	return isgraph(ch);
}

TUCHAR tuchar_tolower( TUCHAR ch )
{
	if (ch >= 0x7f)
		return ch;
	return tolower(ch);
}

TUCHAR tuchar_toupper( TUCHAR ch )
{
	if (ch >= 0x7f)
		return ch;
	return toupper(ch);
}

TUCHAR_TYPELIST tuchar_getchartype( TUCHAR ch )
{
	if (ch >= TUCHAR_WBYTE) {
		/* 2byte */
		if (ch == TUCHAR_WSPACE)
			return TUCHAR_TYPELIST_WIDESPACE;
		else
			return TUCHAR_TYPELIST_WIDECHARS;
	} else {
		/* 1byte */
		if (ch == TUCHAR_SPACE)
			return TUCHAR_TYPELIST_HALFSPACE;
		else
			return TUCHAR_TYPELIST_HALFCHARS;
	}
}

int tuchar_getbytes( TUCHAR ch )
{
	return 2;
}

int tuchar_getbytes_utf8( TUCHAR ch )
{
	if (ch >= 0x800U)
		return 3;
	if (ch >= 0x80U)
		return 2;
	return 1;
}

int tuchar_strlen( const TUCHAR *s )
{
	int ct = 0;

	while (*s) {
		ct++;
		s++;
	}
	return ct;
}

TUCHAR *tuchar_strcpy( TUCHAR *s, const char *cp )
{
	int index = 0;

	while (*cp) {
		s[index] = *cp;
		cp++;
		index++;
	}
	s[index] = '\0';
	return s;
}

TUCHAR *tuchar_strcat( TUCHAR *s, const char *cp )
{
	int index = tuchar_strlen( s );

	while (*cp) {
		s[index] = *cp;
		cp++;
		index++;
	}
	s[index] = '\0';
	return s;
}

TUCHAR *tuchar_chars2tu( iconv_t *id, const char *codename, const char *loc_p, size_t loc_len )
{
#ifdef HAVE_ICONV
	size_t buff_len;
	TUCHAR *tp;
#endif
#if defined(HAVE_ICONV) || defined(HAVE_WCHAR_H)
	size_t uni_len;
	TUCHAR *uni_p;
#endif
	TUCHAR *uni_op;

	if (!*id) {
#ifdef HAVE_ICONV
		if ((*id = iconv_open( UCSNAME, codename )) == (iconv_t)-1)
			ERRMSG("iconv_open() failed, use internal 8-bit through. (%s <- %s)\n", UCSNAME, codename );
#else
		*id = (iconv_t)-1;
#endif
	}
	if (!loc_p)
		loc_len = 0;
	if (loc_len < 0 || loc_len == (size_t)-1)
		loc_len = strlen( loc_p );
#ifdef HAVE_ICONV
	if (!*id || *id == (iconv_t)-1 || !loc_len)
#elif defined(HAVE_WCHAR_H)
	if (!loc_len)
#endif
	{
		size_t i;

		uni_op = (TUCHAR *)TUCHAR_MALLOC( (loc_len + 1) * sizeof(TUCHAR), "tuchar_chars2tu" );
		if (!uni_op) {
			WARNMSG("iconv chars2tu-malloc() failed...\n" );
			return NULL;
		}
		for (i = 0; i < loc_len; i++)
			uni_op[i] = ((TUCHAR)loc_p[i]) & 0x00ffU;
		uni_op[loc_len] = '\0';
	}
#ifdef HAVE_ICONV
	else {
		buff_len = loc_len * sizeof(TUCHAR);
		uni_len = buff_len;
		uni_op = uni_p = TUCHAR_MALLOC( buff_len + sizeof(TUCHAR), "tuchar_chars2tu" );
		if (!uni_op) {
			WARNMSG("iconv chars2tu-malloc() failed...\n" );
			return NULL;
		}
		if (iconv( *id, NULL, NULL, NULL, NULL ) < 0) {
			WARNMSG("iconv chars2tu2 null null null null failed...\n" );
			return NULL;
		}
		while (1) {
			if (iconv( *id, (char **)&loc_p, &loc_len, (char **)&uni_p, &uni_len ) < 0) {
				if (errno == E2BIG) {
					buff_len += 128 * sizeof(TUCHAR);
					uni_len += 128 * sizeof(TUCHAR);
					tp = TUCHAR_REALLOC( uni_op, buff_len + sizeof(TUCHAR), "tuchar_chars2tu" );
					if (tp) {
						uni_p = uni_p - uni_op + tp;
						uni_op = tp;
						continue;
					}
				} else
					WARNMSG("iconv() local code to unicode failed. (%s <- %s)\n", codename, UCSNAME );
				TUCHAR_FREE( uni_op, "tuchar_chars2tu" );
				return NULL;
			}
			break;
		}
		*uni_p = 0;
	}
#elif defined(HAVE_WCHAR_H)
	{
		size_t i;
		wchar_t *wp, *wpp;

		wp = wpp = (wchar_t *)TUCHAR_MALLOC( (loc_len + 1) * 2 * sizeof(wchar_t), "tuchar_chars2tu/wcsbuf" );
		if (!wp) {
			WARNMSG("code-conv chars2tu/wcsbuf-malloc() failed...\n" );
			return NULL;
		}
		mbstowcs( wp, loc_p, (loc_len + 1) * 2 );
		wp[(loc_len + 1) * 2 - 1] = '\0';
		uni_len = wcslen( wp );
		uni_op = uni_p = (TUCHAR *)TUCHAR_MALLOC( (uni_len + 1) * sizeof(TUCHAR), "tuchar_chars2tu/tubuf" );
		if (!uni_op) {
			WARNMSG("code-conv chars2tu/tubuf-malloc() failed...\n" );
			return NULL;
		}
		for (i = 0; i < uni_len; i++)
			*uni_p++ = (TUCHAR)*wpp++;
		*uni_p = 0;
		TUCHAR_FREE( wp, "tuchar_chars2tu/tubuf" );
	}
#endif
	return uni_op;
}

TUCHAR *tuchar_loc2tu( const char *loc_p, size_t loc_len )
{
	return tuchar_chars2tu( &id_l2u, term_codename, loc_p, loc_len );
}

TUCHAR *tuchar_file2tu( const char *loc_p, size_t loc_len )
{
	return tuchar_chars2tu( &id_l2u, file_codename, loc_p, loc_len );
}

TUCHAR *tuchar_utf2tu( const char *loc_p, size_t loc_len )
{
	return tuchar_chars2tu( &id_utf2u, UTF8NAME, loc_p, loc_len );
}

static char *tuchar_tu2chars( iconv_t *id, const char *codename, const TUCHAR *tuchar_p, size_t tuchar_len )
{
#ifdef HAVE_ICONV
	size_t loc_len, buff_len;
	char *loc_p, *tp;
#endif
#if defined(HAVE_ICONV) || defined(HAVE_WCHAR_H)
	size_t byte_len;
#endif
	char *loc_op;

	if (!*id) {
#ifdef HAVE_ICONV
		if ((*id = iconv_open( codename, UCSNAME )) == (iconv_t)-1)
			ERRMSG("iconv_open() failed, use internal 8-bit through. (%s <- %s)\n", codename, UCSNAME );
#else
		*id = (iconv_t)-1;
#endif
	}

	/* convert */
	if (!tuchar_p)
		tuchar_len = 0;
	if (tuchar_len < 0 || tuchar_len == (size_t)-1)
		tuchar_len = tuchar_strlen( tuchar_p );
#ifdef HAVE_ICONV
	if (!*id || *id == (iconv_t)-1 || !tuchar_len)
#elif defined(HAVE_WCHAR_H)
	if (!tuchar_len)
#endif
	{
		size_t i;

		loc_op = (char *)TUCHAR_MALLOC( tuchar_len + 1, "tuchar_tu2chars" );
		if (!loc_op) {
			WARNMSG("iconv tu2chars-malloc() failed...\n" );
			return NULL;
		}
		for (i = 0; i < tuchar_len; i++)
			loc_op[i] = (char)(tuchar_p[i] & 0x00ffU);
		loc_op[tuchar_len] = '\0';
	}
#ifdef HAVE_ICONV
	else {
		byte_len = tuchar_len * sizeof(TUCHAR);
		buff_len = byte_len;
		loc_len = buff_len;
		loc_op = loc_p = TUCHAR_MALLOC( buff_len + 1, "tuchar_tu2chars" );
		if (!loc_op) {
			WARNMSG("iconv tu2chars-malloc() failed...\n" );
			return NULL;
		}
		if (iconv( *id, NULL, NULL, NULL, NULL ) < 0) {
			WARNMSG("iconv tu2chars null null null null failed...\n" );
			return NULL;
		}
		while (1) {
			if (iconv( *id, (char **)&tuchar_p, &byte_len, &loc_p, &loc_len ) < 0) {
				if (errno == E2BIG) {
					buff_len += 128;
					loc_len += 128;
					tp = TUCHAR_REALLOC( loc_op, buff_len + 1, "tuchar_tu2chars" );
					if (tp) {
						loc_p = loc_p - loc_op + tp;
						loc_op = tp;
						continue;
					}
				} else
					WARNMSG("iconv() unicode to local code failed. (%s <- %s)\n", codename, UCSNAME );
				TUCHAR_FREE( loc_op, "tuchar_tu2chars" );
				return NULL;
			}
			break;
		}
		*loc_p = 0;
	}
#elif defined(HAVE_WCHAR_H)
	{
		size_t i;
		wchar_t *wp, *wpp;

		wp = wpp = (wchar_t *)TUCHAR_MALLOC( (tuchar_len + 1) * sizeof(wchar_t), "tuchar_tu2chars/wcsbuf" );
		if (!wp) {
			WARNMSG("code-conv tu2chars/wcsbuf-malloc() failed...\n" );
			return NULL;
		}
		for (i = 0; i < tuchar_len; i++)
			*wpp++ = (wchar_t)*tuchar_p++;
		*wpp = 0;
		loc_op = (char *)TUCHAR_MALLOC( (tuchar_len + 1) * 8, "tuchar_tu2chars/mbsbuf" );
		if (!loc_op) {
			WARNMSG("code-conv tu2chars/mbsbuf-malloc() failed...\n" );
			return NULL;
		}
		wcstombs( loc_op, wp, (tuchar_len + 1) * 8 );
		loc_op[(tuchar_len + 1) * 8 - 1] = '\0';
		TUCHAR_FREE( wp, "tuchar_tu2chars/mbsbuf" );
		byte_len = strlen( loc_op );
		loc_op = (char *)TUCHAR_REALLOC( loc_op, byte_len + 1, "tuchar_tu2chars/mbsbuf" );
		if (!loc_op) {
			WARNMSG("code-conv tu2chars/mbsbuf-realloc() failed...\n" );
			return NULL;
		}
	}
#endif
	return loc_op;
}

char *tuchar_tu2loc( const TUCHAR *tuchar_p, size_t tuchar_len )
{
	return tuchar_tu2chars( &id_u2l, term_codename, tuchar_p, tuchar_len );
}

char *tuchar_tu2file( const TUCHAR *tuchar_p, size_t tuchar_len )
{
	return tuchar_tu2chars( &id_u2l, file_codename, tuchar_p, tuchar_len );
}

char *tuchar_tu2utf( const TUCHAR *tuchar_p, size_t tuchar_len )
{
	return tuchar_tu2chars( &id_u2utf, UTF8NAME, tuchar_p, tuchar_len );
}

int tuchar_fputs( const TUCHAR *tuchar_p, size_t tuchar_len, FILE *fp )
{
	char *r;

	r = tuchar_tu2loc( tuchar_p, tuchar_len );
	fputs( r, fp );
	TUCHAR_FREE( r, "tuchar_fputs" );
	return 0;
}

static void tuchar_close_iconv( void )
{
	if (id_l2u != NULL && id_l2u != (iconv_t)-1) {
#ifdef HAVE_ICONV
		iconv_close( id_l2u );
#endif
		id_l2u = NULL;
	}
	if (id_u2l != NULL && id_u2l != (iconv_t)-1) {
#ifdef HAVE_ICONV
		iconv_close( id_u2l );
#endif
		id_u2l = NULL;
	}
	if (id_f2u != NULL && id_f2u != (iconv_t)-1) {
#ifdef HAVE_ICONV
		iconv_close( id_f2u );
#endif
		id_f2u = NULL;
	}
	if (id_u2f != NULL && id_u2f != (iconv_t)-1) {
#ifdef HAVE_ICONV
		iconv_close( id_u2f );
#endif
		id_u2f = NULL;
	}
	if (id_utf2u != NULL && id_utf2u != (iconv_t)-1) {
#ifdef HAVE_ICONV
		iconv_close( id_utf2u );
#endif
		id_utf2u = NULL;
	}
	if (id_u2utf != NULL && id_u2utf != (iconv_t)-1) {
#ifdef HAVE_ICONV
		iconv_close( id_u2utf );
#endif
		id_u2utf = NULL;
	}
	return;
}

int tuchar_setcodingname( const char *name, int use_external_codename_buf )
{
	if (!name)
		term_codename = term_codename_default;
	else {
		if (!use_external_codename_buf) {
			strncpy( term_codename_buf, name, LNMAX - 1 );
			term_codename_buf[LNMAX - 1] = '\0';
			term_codename = term_codename_buf;
		} else
			term_codename = name;
	}
#ifdef HAVE_ICONV
	tuchar_close_iconv();
#elif defined(HAVE_WCHAR_H)
#ifdef HAVE_LOCALE_H
	setlocale( LC_ALL, term_codename );
#endif
#endif /* ICONV or WCHAR */
	return 0;
}

int tuchar_setfilecodingname( const char *name, int use_external_codename_buf )
{
	if (!name)
		file_codename = term_codename_default;
	else {
		if (!use_external_codename_buf) {
			strncpy( file_codename_buf, name, LNMAX - 1 );
			file_codename_buf[LNMAX - 1] = '\0';
			file_codename = file_codename_buf;
		} else
			file_codename = name;
	}
#ifdef HAVE_ICONV
	tuchar_close_iconv();
#endif /* ICONV */
	return 0;
}

const char *tuchar_getcodingname( void )
{
	return term_codename;
}

const char *tuchar_getfilecodingname( void )
{
	return file_codename;
}

int tuchar_init( const char *locstr )
{
#ifdef HAVE_ICONV
	const char *cp = NULL;
	const char *langstr;
#endif

#ifdef TUCHAR_DISP_TERMINALMODE
	if (locstr)
		printf("locstr:%s\n", locstr );
	else
		printf("locstr:NULL\n" );
#endif
	id_l2u = NULL;
	id_u2l = NULL;
	id_f2u = NULL;
	id_u2f = NULL;
	id_utf2u = NULL;
	id_u2utf = NULL;

#ifdef HAVE_ICONV
	if (locstr && locstr[0]) {
		cp = strrchr( locstr, '.' );
		if (cp)
			cp++;
	}
	if (!cp || !locstr || strcasecmp( locstr, "C" )) {
		langstr = getenv( "LANG" );
#ifdef TUCHAR_DISP_TERMINALMODE
		if (langstr)
			printf("langstr:%s\n", langstr );
		else
			printf("langstr:NULL\n" );
#endif
		if (langstr && langstr[0]) {
			cp = strrchr( langstr, '.' );
			if (cp)
				cp++;
		}
	}
	if (!cp) {
		if (locstr) {
			if (strncasecmp( locstr, "ja", 2 ))
				cp = "ISO 8859-1";
			else
				cp = "utf8";
		} else if (langstr) {
			if (strncasecmp( langstr, "ja_JP", 5 ))
				cp = "ISO 8859-1";
			else
				cp = "utf8";
		} else
			cp = "ISO 8859-1";
	}
	strcpy( term_codename_default, cp );
#elif defined(HAVE_WCHAR_H)
	/* Currently Support for WIN32 only */
	strcpy( term_codename_default, locstr );
#else
	strcpy( term_codename_default, "" );
#endif
	term_codename = term_codename_default;
	file_codename = term_codename_default;
#ifdef TUCHAR_DISP_TERMINALMODE
	printf("%s\n", term_codename );
#endif
	return TUCHAR_RESULT_NOERROR;
}

void tuchar_shutdown( void )
{
	tuchar_close_iconv();
	return;
}

/*
	Local Variables:
	mode:c++
	c-set-style:"k&r"
	c-basic-offset:8
	tab-width:8
	End:
*/
