/* File: bmp2pat.c

	Tool for MZ-2000/2200/80B/B2 Emulator "emz2000"
		EmuZ-2000 to emz2000 font converter

	Copyright (C) 2002-2016 Toshio Fukui
	This file is part of the emz2000 software.
	See copyright notice in the COPYING file.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#include "bmpstruct.h"

#define TRUE	1
#define FALSE	0
#define LNMAX	256

#define BMP2PAT_VERSION		"1.14"
#define BMP2PAT_COPYRIGHT	"2002-2016"

char const *_copyright =
	"EmuZ-2000 font flie (.bmp) to .pat Ver." BMP2PAT_VERSION "\n"
	"Copyright (C) " BMP2PAT_COPYRIGHT " Toshio Fukui.\n";

char module_name[LNMAX];

int bmptopat( char *from_fn, char *to_fn, int font16_flag )
{
	int i, j, k, col_bit, col_size;
	int tflag, r;
	unsigned char c, *p, *buf;
	unsigned char *im_p = NULL, *bitdata_p;
	FILE *fp;
	BITMAPFILEHEADER *filehead_p;
	BITMAPINFOHEADER *fileinfo_p;
	RGBQUAD *rgbcol_p;

	/* open the BMP file */
	fp = fopen( from_fn, "rb" );
	if (!fp) {
		fprintf(stderr, "can't read open the BMP file \"%s\"\n", from_fn);
		return FALSE;
	}

	/* headder */
	buf = malloc(sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER));
	if (!buf) {
		fprintf(stderr, "can't allocate memory.\n");
		fclose(fp);
		return FALSE;
	}
	r = fread( buf, sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER), 1, fp );
	if (r != 1) {
		fprintf(stderr, "read failed the BMP file - 1 \"%s\"\n", from_fn);
		free(buf);
		fclose(fp);
		return FALSE;
	}
	filehead_p = (BITMAPFILEHEADER *)buf;
	fileinfo_p = (BITMAPINFOHEADER *)(buf + sizeof(BITMAPFILEHEADER));

	/* check bitmap type */
	if (strncmp( (char *)(filehead_p -> bfType), "BM", 2) ) {
		fprintf(stderr, "not supportted bitmap type (eyecatch != BM)\n");
		free(buf);
		fclose(fp);
		return FALSE;
	}
	if (fileinfo_p -> biPlanes != 1) {
		fprintf(stderr, "not supportted bitmap type (biPlanes != 1)\n");
		free(buf);
		fclose(fp);
		return FALSE;
	}
	if (fileinfo_p -> biCompression != BI_RGB) {
		fprintf(stderr, "not supportted bitmap type (biCompression != BI_RGB)\n");
		free(buf);
		fclose(fp);
		return FALSE;
	}
	if (fileinfo_p -> biCompression == BI_BITFIELDS && fileinfo_p -> biBitCount != 16) {
		fprintf(stderr, "not supportted bitmap type (biCompression != BI_RGB && BitCount != 16)\n");
		free(buf);
		fclose(fp);
		return FALSE;
	}
	col_bit = fileinfo_p -> biBitCount;
	switch (col_bit) {
	case 1 :
		if (fileinfo_p -> biClrUsed)
			col_size = fileinfo_p -> biClrUsed;
		else
			col_size = 2;
		break;
	default :
		fprintf(stderr, "not supportted bitmap type (biBitcount != 1)\n");
		free(buf);
		fclose(fp);
		return FALSE;
	}

	/* read RGB info */
	if (!(rgbcol_p = malloc( sizeof(RGBQUAD)*col_size ))) {
		fprintf(stderr, "can't allocate memory. (colormap-1)\n");
		free(buf);
		fclose(fp);
		return FALSE;
	}
	r = fread( rgbcol_p, sizeof(RGBQUAD)*col_size, 1, fp );
	if (r != 1) {
		fprintf(stderr, "read failed the BMP file - 2 \"%s\"\n", from_fn);
		free(buf);
		fclose(fp);
		return FALSE;
	}

	/* read bit table */
	bitdata_p = malloc( col_bit*((fileinfo_p -> biWidth + 7)/8)*(fileinfo_p -> biHeight) );
	if (!bitdata_p) {
		fprintf(stderr, "can't allocate memory. (image temporary).\n");
		free(buf);
		fclose(fp);
		return FALSE;
	}
	memset( bitdata_p, 0, col_bit*((fileinfo_p -> biWidth + 7)/8)*(fileinfo_p -> biHeight) );
	r = fread( bitdata_p, 1, col_bit*((fileinfo_p -> biWidth + 7)/8)*(fileinfo_p -> biHeight), fp );
	if (r != col_bit*((fileinfo_p -> biWidth + 7)/8)*(fileinfo_p -> biHeight)) {
		fprintf(stderr, "read failed the BMP file - 3 \"%s\"\n", from_fn);
		free(buf);
		free(bitdata_p);
		fclose(fp);
		return FALSE;
	}

	/* read data (allocate) */
	p = bitdata_p;
	im_p = malloc( (fileinfo_p -> biWidth) * (fileinfo_p -> biHeight) );
	if (!im_p) {
		fprintf(stderr, "can't allocate memory. (image)\n");
		free(buf);
		free(bitdata_p);
		fclose(fp);
		return FALSE;
	}
	for (i = fileinfo_p -> biHeight - 1; i >= 0; --i) {
		tflag = TRUE;
		for (j = 0; j < fileinfo_p -> biWidth; j++) {
			im_p[(i*(fileinfo_p -> biWidth)+j)] = (*p >> (7-j % 8)) & 0x01;
			if (j % 8 == 7) {
				p++;
				tflag = TRUE;
			} else
				tflag = FALSE;
		}
		if (!tflag)
			p++;
		while (((unsigned long)p-(unsigned long)bitdata_p) & 0x03)
			p++;
	}

	/* post processing */
//	free(buf);			// for gcc-5.3
//	free(bitdata_p);		// for gcc-5.3
	fclose(fp);

	/* create PAT file */
	fp = fopen(to_fn, "w");
	if (!fp) {
		fprintf(stderr, "can't write open the PAT file \"%s\"\n", to_fn);
		free(im_p);
		return FALSE;
	}
	for (i = 0; i < 256; i++) {
		fprintf(fp, "# %d\n", i);
		for (j = 0; j < 8; j++) {
			for (k = 0; k < 8; k++) {
				if (font16_flag) {
					c = !im_p[(fileinfo_p -> biWidth)*((i % 16)*16+j*2)+k+(i / 16)*8];
					c |= !im_p[(fileinfo_p -> biWidth)*((i % 16)*16+j*2+1)+k+(i / 16)*8];
				} else
					c = !im_p[(fileinfo_p -> biWidth)*((i % 16)*8+j)+k+(i / 16)*8];
				fputc( c ? 'X' : '.', fp );
			}
			fputc( '\n', fp );
		}
		fputc( '\n', fp );
	}
	free(im_p);
	free(buf);			// for gcc-5.3
	free(bitdata_p);		// for gcc-5.3
	fclose(fp);
	return TRUE;
}

int usage( void )
{
	fprintf( stderr, "%s\n", _copyright );
	fprintf( stderr, "Usage: %s [optins] <input filename> <output filename>\n", module_name );
	fprintf( stderr, "\t-x ... select the 16 dot font\n");
	fprintf( stderr, "\t-v ... display version information\n\n" );
	return 10;
}

int main( int argc, char **argv )
{
	int r, font16_flag;
	char *optstring = "xv";

	strncpy( module_name, argv[0], LNMAX );
	module_name[LNMAX - 1] = '\0';
	font16_flag = FALSE;
	while (1) {
		int c;

		c = getopt( argc, argv, optstring );
		if (c == -1)
			break;
		switch (c) {
		case 'x' :
			font16_flag = TRUE;
			break;
		case 'v' :
			fprintf( stderr, "%s", _copyright );
			return 10;
		default :
			return usage();
		}
	}

	if (argc - optind != 2)
		return usage();
	if (!bmptopat( argv[optind], argv[optind + 1], font16_flag )) {
		fprintf( stderr, "*** Convert error ***\n" );
		unlink( argv[2] );
		r = 1;
	} else {
		fprintf( stderr, "*** Completed ***\n" );
		r = 0;
	}
	return r;
}

/*
	Local Variables:
	mode:c++
	c-set-style:"k&r"
	c-basic-offset:8
	tab-width:8
	End:
*/
